import asyncio
import sys

from .config import CLIConfig, LoadConfigError, PyprojectConfig, load_config
from .logging import logging_context
from .runner import TaskRunner
from .utils import catch_unknown_errors, killed_by


@catch_unknown_errors()
@killed_by(LoadConfigError)
def main() -> None:
    config = load_config(CLIConfig, PyprojectConfig)
    with logging_context(config.log_levels):
        success = asyncio.run(TaskRunner(config.tasks).run_tasks())
    sys.exit(not success)



from .config import CLIConfig, Config, ConfigLoader, LoadConfigError, PyprojectConfig, load_config
from .logging import LogLevel, LogLevels, logging_context
from .runner import Task, TaskRunner

__all__ = [
    'CLIConfig',
    'Config',
    'ConfigLoader',
    'LoadConfigError',
    'LogLevel',
    'LogLevels',
    'PyprojectConfig',
    'Task',
    'TaskRunner',
    'load_config',
    'logging_context',
]



import logging
from dataclasses import dataclass

from .formatting import FAIL, bright, dim, status
from .logging import ASYNC_LOG, task_log
from .utils import concurrent_call, exec_command, timed_awaitable

log = ASYNC_LOG


@dataclass
class Task:
    command: str
    name: str = ''

    @property
    def id(self) -> str:
        return self.name or self.command


class TaskRunner:
    def __init__(self, tasks: list[Task]) -> None:
        self.tasks = tasks
        self.has_named_tasks = any(t.name for t in tasks)
        self.max_name_length = max(len(t.id) for t in tasks or [Task('')])

    async def run_tasks(self) -> bool:
        if not self.tasks:
            log.warning('Nothing to do. Getting bored...\n')
            return True
        if self.has_named_tasks:
            await self._show_todo()
        results = await concurrent_call(self._run_task, self.tasks)
        failed_tasks = [task for task, ok in results if not ok]
        if failed_tasks:
            log.error('')
            log.error(f'{FAIL} {bright("Failed tasks:")} {failed_tasks}')
        return not failed_tasks

    def _task_line(self, bullet: str, task: Task, data: str) -> str:
        return f'{bullet} {task.id.ljust(self.max_name_length)}  {dim(data)}'

    async def _show_todo(self) -> None:
        summary = [self._task_line('•', task, task.command) for task in self.tasks]
        for line in (bright('To do:'), *summary, '', bright('Results:')):
            log.info(line)

    async def _run_task(self, task: Task) -> tuple[str, bool]:
        (success, output_streams), duration = await timed_awaitable(exec_command(task.command))
        log_level = logging.INFO if success else logging.ERROR
        log.log(log_level, self._task_line(status(success), task, duration))
        for stream in output_streams:
            if stream:
                task_log(success).log(log_level, stream.decode())
        return task.id, success



import logging
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import asdict, dataclass
from enum import IntEnum
from logging.handlers import QueueHandler, QueueListener
from multiprocessing import Queue

from gaffe import raises

ASYNC_LOG = logging.getLogger('powerchord.all')


def task_log(success: bool) -> logging.Logger:
    return logging.getLogger('powerchord.' + ('success' if success else 'fail'))


class LogLevel(IntEnum):
    NEVER = 100
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG

    @classmethod
    @raises(ValueError)
    def decode(cls, value: str) -> 'LogLevel':
        if not value:
            return cls.NEVER
        try:
            return cls[value.upper()]
        except (AttributeError, KeyError) as exc:
            raise ValueError('Invalid log level:', value) from exc


@dataclass
class LogLevels:
    all: LogLevel = LogLevel.INFO
    success: LogLevel = LogLevel.NEVER
    fail: LogLevel = LogLevel.ERROR


def setup_logging_queues(levels: LogLevels) -> Iterator[QueueListener]:
    console = logging.StreamHandler(sys.stdout)
    logging.basicConfig(handlers=[console], level=levels.all, format='%(message)s')
    queue: Queue[logging.LogRecord] = Queue()
    for name, level in asdict(levels).items():
        logger = logging.getLogger('powerchord.' + name)
        logger.setLevel(max(level, levels.all))
        logger.addHandler(QueueHandler(queue))
        logger.propagate = False
    yield QueueListener(queue, console)


@contextmanager
def logging_context(levels: LogLevels) -> Iterator[None]:
    queues_listeners = list(setup_logging_queues(levels))
    for listener in queues_listeners:
        listener.start()
    try:
        yield
    finally:
        for listener in queues_listeners:
            listener.stop()



import argparse
import tomllib
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar

from chili import decode
from gaffe import raises

from .logging import LogLevel, LogLevels
from .runner import Task


class ParseConfigError(Exception):
    pass


@dataclass
class Config:
    tasks: list[Task] = field(default_factory=list)
    log_levels: LogLevels = field(default_factory=LogLevels)


class DecodeConfigError(Exception):
    pass


@dataclass
class ConfigLoader(ABC):
    name: ClassVar[str]

    tasks: list[Task] = field(default_factory=list)
    log_levels: LogLevels = field(default_factory=LogLevels)

    @classmethod
    @raises(ParseConfigError)
    @abstractmethod
    def _parse(cls) -> dict:
        pass

    @classmethod
    @raises(ParseConfigError, DecodeConfigError)
    def load(cls) -> Config | None:
        config_dict = cls._parse()
        if not any(config_dict.values()):
            return None

        tasks = config_dict.get('tasks', {})
        if isinstance(tasks, list):
            task_items = [('', t) if isinstance(t, str) else t for t in tasks]
        elif isinstance(tasks, dict):
            task_items = list(tasks.items())
        else:
            raise DecodeConfigError(f'Wrong value for tasks: {tasks}')
        config_dict['tasks'] = [{'command': t, 'name': n} for n, t in task_items]

        log_levels = config_dict.get('log_levels', {})
        if isinstance(log_levels, list):
            config_dict['log_levels'] = dict(log_levels)

        try:
            return decode(config_dict, Config, decoders={LogLevel: LogLevel})
        except ValueError as exc:
            raise DecodeConfigError(*exc.args) from exc


def parse_key_value_pair(value: str) -> tuple[str, str]:
    key, value = value.split('=', 1)
    return key, value


def try_parse_key_value_pair(value: str) -> str | tuple[str, str]:
    try:
        return parse_key_value_pair(value)
    except ValueError:
        return value


class CLIConfig(ConfigLoader):
    name = 'command line'

    @classmethod
    @raises(ParseConfigError)
    def _parse(cls) -> dict:
        arg_parser = argparse.ArgumentParser()
        arg_parser.add_argument(
            '-t',
            '--tasks',
            dest='tasks',
            nargs='+',
            metavar='COMMAND | NAME=COMMAND',
            type=try_parse_key_value_pair,
            default={},
        )
        arg_parser.add_argument(
            '-l',
            '--log-levels',
            dest='log_levels',
            nargs='+',
            metavar='OUTPUT=LOGLEVEL (debug | info | warning | error | critical | "")',
            type=parse_key_value_pair,
            default={},
        )
        try:
            return arg_parser.parse_args().__dict__
        except (argparse.ArgumentError, SystemExit, TypeError) as exc:
            raise ParseConfigError from exc


class PyprojectConfig(ConfigLoader):
    name = 'pyproject.toml'

    @classmethod
    @raises(ParseConfigError)
    def _parse(cls) -> dict:
        try:
            with Path('pyproject.toml').open('rb') as f:
                return tomllib.load(f).get('tool', {}).get('powerchord', {})
        except OSError:
            return {}
        except ValueError as exc:
            raise ParseConfigError from exc


class LoadConfigError(Exception):
    def __init__(self, name: str = '', *args):
        super().__init__(f'Could not load config{name}{":" if args else ""}', *args)


class LoadSpecificConfigError(LoadConfigError):
    def __init__(self, name: str, *args):
        super().__init__(f' from {name}', *args)


@raises(LoadConfigError)
def load_config(*loaders: type[ConfigLoader]) -> Config:
    for loader_cls in loaders:
        try:
            config = loader_cls.load()
        except (ParseConfigError, DecodeConfigError) as exc:
            raise LoadSpecificConfigError(loader_cls.name, *exc.args) from exc
        if config:
            return config
    raise LoadConfigError



import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine, Iterable
from functools import wraps
from subprocess import PIPE
from time import perf_counter_ns
from typing import Any


class FatalError(SystemExit):
    def __init__(self, *args):
        super().__init__(' '.join(str(a) for a in ['💀', *args]))


def killed_by[E: Exception, **P, T](
        *errors: type[E],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return func(*args, **kwargs)
            except errors as exc:
                raise FatalError(*exc.args) from exc
        return wrapper
    return decorator


def catch_unknown_errors[**P, T](
    unknown_message: str = 'Unknown error',
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return func(*args, **kwargs)
            except Exception as exc:
                raise FatalError(unknown_message) from exc
        return wrapper
    return decorator


async def exec_command(command: str) -> tuple[bool, tuple[bytes, bytes]]:
    proc = await asyncio.create_subprocess_shell(command, stdout=PIPE, stderr=PIPE)
    output_streams = await proc.communicate()
    return proc.returncode == 0, output_streams


async def concurrent_iter[T](
    coroutines: Iterable[Coroutine[Any, Any, T]],
) -> AsyncIterator[T]:
    tasks: list[asyncio.Task[T]] = [asyncio.create_task(coro) for coro in coroutines]
    for task in tasks:
        yield await task


async def concurrent_list[T](
    coroutines: Iterable[Coroutine[Any, Any, T]],
) -> list[T]:
    return [item async for item in concurrent_iter(coroutines)]


async def concurrent_call[A, T](
    async_func: Callable[[A], Coroutine[Any, Any, T]],
    args_list: list[A],
) -> list[T]:
    return await concurrent_list(async_func(args) for args in args_list)


def human_readable_duration(nanoseconds: int) -> str:
    minutes = int(nanoseconds // 60_000_000_000)
    nanoseconds %= 60_000_000_000
    seconds = int(nanoseconds // 1_000_000_000)
    nanoseconds %= 1_000_000_000
    milliseconds = int(nanoseconds // 1_000_000)
    nanoseconds %= 1_000_000
    microseconds = int(nanoseconds // 1_000)
    nanoseconds %= 1_000
    if minutes:
        return f'{minutes:d}:{seconds:02d}.{milliseconds:03d} minutes'
    if seconds:
        return f'{seconds:d}.{milliseconds:03d} seconds'
    if milliseconds:
        return f'{milliseconds:d}.{microseconds:03d} ms'
    return f'{microseconds:d}.{nanoseconds:03d} µs'


def timed[T](
    func: Callable[[], T],
    formatter: Callable[[int], str] = None,
) -> tuple[T, str]:
    start = perf_counter_ns()
    return func(), (formatter or human_readable_duration)(perf_counter_ns() - start)


def timed_awaitable[T](
    awaitable: Awaitable[T],
    formatter: Callable[[int], str] = None,
) -> Awaitable[tuple[T, str]]:
    async def wrapper() -> tuple[T, str]:
        start = perf_counter_ns()
        return await awaitable, (formatter or human_readable_duration)(perf_counter_ns() - start)
    return wrapper()
